import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Tuple
from . import DEFAULT_TIME # Added DEFAULT_TIME

# Use Rich for pretty printing to the console
from rich.console import Console
# Initialize Rich Console
console = Console(record=True)
rprint = console.print

# Use relative import for internal modules
try:
    # Attempt relative import for package context
    from .fix_code_module_errors import fix_code_module_errors
except ImportError:
    # Fallback for script execution context (e.g., testing)
    # This assumes fix_code_module_errors.py is in the same directory or Python path
    # You might need to adjust this based on your project structure during testing
    print("Warning: Relative import failed. Attempting direct import for fix_code_module_errors.", file=sys.stderr)
    # Add parent directory to sys.path if necessary for testing outside a package
    # import sys
    # sys.path.append(str(Path(__file__).parent.parent)) # Adjust based on structure
    from fix_code_module_errors import fix_code_module_errors


def fix_code_loop(
    code_file: str,
    prompt: str,
    verification_program: str,
    strength: float,
    temperature: float,
    max_attempts: int,
    budget: float,
    error_log_file: str,
    verbose: bool = False,
    time: float = DEFAULT_TIME,
) -> Tuple[bool, str, str, int, float, str | None]:
    """
    Attempts to fix errors in a code module through multiple iterations.

    Args:
        code_file: Path to the code file being tested.
        prompt: The prompt that generated the code under test.
        verification_program: Path to the Python program that verifies the code.
        strength: LLM model strength (0.0 to 1.0).
        temperature: LLM temperature (0.0 to 1.0).
        max_attempts: Maximum number of fix attempts.
        budget: Maximum cost allowed for the fixing process.
        error_log_file: Path to the error log file.
        verbose: Enable detailed logging (default: False).
        time: Time limit for the LLM calls (default: DEFAULT_TIME).

    Returns:
        Tuple containing the following in order:
        - success (bool): Whether the errors were successfully fixed.
        - final_program (str): Contents of the final verification program file (empty string if unsuccessful).
        - final_code (str): Contents of the final code file (empty string if unsuccessful).
        - total_attempts (int): Number of fix attempts made.
        - total_cost (float): Total cost of all fix attempts.
        - model_name (str | None): Name of the LLM model used (or None if no LLM calls were made).
    """
    # --- Start: Modified File Checks ---
    if not Path(code_file).is_file():
        # Raising error for code file is acceptable as it's fundamental
        raise FileNotFoundError(f"Code file not found: {code_file}")
    if not Path(verification_program).is_file():
        # Handle missing verification program gracefully as per test expectation
        rprint(f"[bold red]Error: Verification program not found: {verification_program}[/bold red]")
        return False, "", "", 0, 0.0, None
    # --- End: Modified File Checks ---

    # Step 1: Remove existing error log file
    try:
        os.remove(error_log_file)
        if verbose:
            rprint(f"Removed existing error log file: {error_log_file}")
    except FileNotFoundError:
        if verbose:
            rprint(f"Error log file not found, no need to remove: {error_log_file}")
    except OSError as e:
        rprint(f"[bold red]Error removing log file {error_log_file}: {e}[/bold red]")
        # Decide if this is fatal or not; for now, we continue

    # Step 2: Initialize variables
    attempts = 0
    total_cost = 0.0
    success = False
    model_name = None
    history_log = "<history>\n" # Initialize history log XML root

    # Create initial backups before any modifications
    code_file_path = Path(code_file)
    verification_program_path = Path(verification_program)
    original_code_backup = f"{code_file_path.stem}_original_backup{code_file_path.suffix}"
    original_program_backup = f"{verification_program_path.stem}_original_backup{verification_program_path.suffix}"

    try:
        shutil.copy2(code_file, original_code_backup)
        shutil.copy2(verification_program, original_program_backup)
        if verbose:
            rprint(f"Created initial backups: {original_code_backup}, {original_program_backup}")
    except Exception as e:
        rprint(f"[bold red]Error creating initial backups: {e}[/bold red]")
        # If backups fail, we cannot guarantee restoration. Return failure.
        return False, "", "", 0, 0.0, None


    # Step 3: Enter the fixing loop
    while attempts < max_attempts and total_cost <= budget:
        current_attempt = attempts + 1 # User-facing attempt number (starts at 1)
        rprint(f"\n[bold cyan]Attempt {current_attempt}/{max_attempts}...[/bold cyan]")
        attempt_log_entry = f'  <attempt number="{current_attempt}">\n' # Start XML for this attempt

        # b. Run the verification program
        if verbose:
            rprint(f"Running verification: {sys.executable} {verification_program}")

        process = subprocess.run(
            [sys.executable, verification_program],
            capture_output=True,
            text=True,
            encoding='utf-8', # Ensure consistent encoding
        )

        verification_status = f"Success (Return Code: {process.returncode})" if process.returncode == 0 else f"Failure (Return Code: {process.returncode})"
        verification_output = process.stdout or "[No standard output]"
        verification_error = process.stderr or "[No standard error]"

        # Add verification results to the attempt log entry
        attempt_log_entry += f"""\
    <verification>
      <status>{verification_status}</status>
      <output><![CDATA[
{verification_output}
]]></output>
      <error><![CDATA[
{verification_error}
]]></error>
    </verification>
"""

        # c. If the program runs without errors, break the loop
        if process.returncode == 0:
            rprint("[bold green]Verification successful![/bold green]")
            success = True
            history_log += attempt_log_entry + "  </attempt>\n" # Close the final successful attempt
            break

        # d. If the program fails
        rprint(f"[bold red]Verification failed with return code {process.returncode}.[/bold red]")
        current_error_message = verification_error # Use stderr as the primary error source

        # Add current error to the attempt log entry
        attempt_log_entry += f"""\
    <current_error><![CDATA[
{current_error_message}
]]></current_error>
"""

        # Check budget *before* making the potentially expensive LLM call for the next attempt
        # (Only check if cost > 0 to avoid breaking before first attempt if budget is 0)
        if total_cost > budget and attempts > 0: # Check after first attempt cost is added
             rprint(f"[bold yellow]Budget exceeded (${total_cost:.4f} > ${budget:.4f}) before attempt {current_attempt}. Stopping.[/bold yellow]")
             history_log += attempt_log_entry + "    <error>Budget exceeded before LLM call</error>\n  </attempt>\n"
             break

        # Check max attempts *before* the LLM call for this attempt
        if attempts >= max_attempts:
             rprint(f"[bold red]Maximum attempts ({max_attempts}) reached before attempt {current_attempt}. Stopping.[/bold red]")
             # No need to add to history here, loop condition handles it
             break


        # Create backup copies for this iteration BEFORE calling LLM
        code_base, code_ext = os.path.splitext(code_file)
        program_base, program_ext = os.path.splitext(verification_program)
        code_backup_path = f"{code_base}_{current_attempt}{code_ext}"
        program_backup_path = f"{program_base}_{current_attempt}{program_ext}"

        try:
            shutil.copy2(code_file, code_backup_path)
            shutil.copy2(verification_program, program_backup_path)
            if verbose:
                rprint(f"Created backups for attempt {current_attempt}: {code_backup_path}, {program_backup_path}")
        except Exception as e:
            rprint(f"[bold red]Error creating backups for attempt {current_attempt}: {e}[/bold red]")
            history_log += attempt_log_entry + f"    <error>Failed to create backups: {e}</error>\n  </attempt>\n"
            break # Cannot proceed reliably without backups

        # Read current file contents
        try:
            current_code = Path(code_file).read_text(encoding='utf-8')
            current_program = Path(verification_program).read_text(encoding='utf-8')
        except Exception as e:
            rprint(f"[bold red]Error reading source files: {e}[/bold red]")
            history_log += attempt_log_entry + "    <error>Failed to read source files</error>\n  </attempt>\n"
            break # Cannot proceed without file contents

        # Prepare the full history context for the LLM
        # Temporarily close the XML structure for the LLM call
        error_context_for_llm = history_log + attempt_log_entry + "  </attempt>\n</history>\n"

        # Call fix_code_module_errors
        rprint("Attempting to fix errors using LLM...")
        update_program, update_code, fixed_program, fixed_code = False, False, "", ""
        program_code_fix, cost, model_name_iter = "", 0.0, None

        # Capture Rich output from the internal function if needed, though it prints directly
        # Using a temporary console or redirect might be complex if it uses the global console
        # For simplicity, we assume fix_code_module_errors prints directly using `rprint`

        try:
            # Note: The example signature for fix_code_module_errors returns 7 values
            (update_program, update_code, fixed_program, fixed_code,
             program_code_fix, cost, model_name_iter) = fix_code_module_errors(
                program=current_program,
                prompt=prompt,
                code=current_code,
                errors=error_context_for_llm, # Pass the structured history
                strength=strength,
                temperature=temperature,
                time=time, # Pass time
                verbose=verbose,
            )
            if model_name_iter:
                 model_name = model_name_iter # Update model name if returned

        except Exception as e:
            rprint(f"[bold red]Error calling fix_code_module_errors: {e}[/bold red]")
            cost = 0.0 # Assume no cost if the call failed
            # Add error to the attempt log entry
            attempt_log_entry += f"""\
    <fixing>
      <error>LLM call failed: {e}</error>
    </fixing>
"""
            # Continue to the next attempt or break if limits reached? Let's break.
            history_log += attempt_log_entry + "  </attempt>\n" # Log the attempt with the LLM error
            break # Stop if the fixing mechanism itself fails

        # Add fixing results to the attempt log entry
        attempt_log_entry += f"""\
    <fixing>
      <llm_analysis><![CDATA[
{program_code_fix or "[No analysis provided]"}
]]></llm_analysis>
      <decision>
        update_program: {str(update_program).lower()}
        update_code: {str(update_code).lower()}
      </decision>
      <cost>{cost:.4f}</cost>
      <model>{model_name_iter or 'N/A'}</model>
    </fixing>
"""
        # Close the XML tag for this attempt
        attempt_log_entry += "  </attempt>\n"
        # Append this attempt's full log to the main history
        history_log += attempt_log_entry

        # Write the cumulative history log to the file *after* each attempt
        try:
            with open(error_log_file, "w", encoding="utf-8") as f:
                f.write(history_log + "</history>\n") # Write complete history including root close tag
        except IOError as e:
            rprint(f"[bold red]Error writing to log file {error_log_file}: {e}[/bold red]")


        # Add cost and check budget *after* the LLM call
        total_cost += cost
        rprint(f"Attempt Cost: ${cost:.4f}, Total Cost: ${total_cost:.4f}, Budget: ${budget:.4f}")
        if total_cost > budget:
            rprint(f"[bold yellow]Budget exceeded (${total_cost:.4f} > ${budget:.4f}) after attempt {current_attempt}. Stopping.[/bold yellow]")
            break # Stop loop

        # If LLM suggested no changes but verification failed, stop to prevent loops
        if not update_program and not update_code and process.returncode != 0:
             rprint("[bold yellow]LLM indicated no changes needed, but verification still fails. Stopping.[/bold yellow]")
             success = False # Ensure success is False
             break # Stop loop

        # Apply fixes if suggested
        try:
            if update_code:
                Path(code_file).write_text(fixed_code, encoding='utf-8')
                rprint(f"[green]Updated code file: {code_file}[/green]")
            if update_program:
                Path(verification_program).write_text(fixed_program, encoding='utf-8')
                rprint(f"[green]Updated verification program: {verification_program}[/green]")
        except IOError as e:
            rprint(f"[bold red]Error writing updated files: {e}[/bold red]")
            success = False # Mark as failed if we can't write updates
            break # Stop if we cannot apply fixes

        # e. Increment attempt counter (used for loop condition)
        attempts += 1

        # Check if max attempts reached after incrementing (for the next loop iteration check)
        if attempts >= max_attempts:
             rprint(f"[bold red]Maximum attempts ({max_attempts}) reached. Final verification pending.[/bold red]")
             # Loop will terminate naturally on the next iteration's check


    # Step 4: Restore original files if the process failed overall
    if not success:
        rprint("[bold yellow]Attempting to restore original files as the process did not succeed.[/bold yellow]")
        try:
            # Check if backup files exist before attempting to restore
            if Path(original_code_backup).exists() and Path(original_program_backup).exists():
                shutil.copy2(original_code_backup, code_file)
                shutil.copy2(original_program_backup, verification_program)
                rprint(f"Restored {code_file} and {verification_program} from initial backups.")
            else:
                rprint(f"[bold red]Error: Initial backup files not found. Cannot restore original state.[/bold red]")
        except Exception as e:
            rprint(f"[bold red]Error restoring original files: {e}. Final files might be in a failed state.[/bold red]")

    # Clean up initial backup files regardless of success/failure
    try:
        if Path(original_code_backup).exists():
             os.remove(original_code_backup)
        if Path(original_program_backup).exists():
             os.remove(original_program_backup)
        if verbose:
            rprint(f"Removed initial backup files (if they existed).")
    except OSError as e:
        rprint(f"[bold yellow]Warning: Could not remove initial backup files: {e}[/bold yellow]")


    # Step 5: Read final file contents and determine return values
    final_code_content = ""
    final_program_content = ""
    # --- Start: Modified Final Content Reading ---
    if success:
        try:
            final_code_content = Path(code_file).read_text(encoding='utf-8')
            final_program_content = Path(verification_program).read_text(encoding='utf-8')
        except Exception as e:
            rprint(f"[bold red]Error reading final file contents even after success: {e}[/bold red]")
            # If we succeeded but can't read files, something is wrong. Mark as failure.
            success = False
            final_code_content = ""
            final_program_content = ""
    else:
        # If not successful, return empty strings as per test expectations
        final_code_content = ""
        final_program_content = ""
    # --- End: Modified Final Content Reading ---

    # Ensure the final history log file is complete
    try:
        with open(error_log_file, "w", encoding="utf-8") as f:
             f.write(history_log + "</history>\n")
    except IOError as e:
        rprint(f"[bold red]Final write to log file {error_log_file} failed: {e}[/bold red]")

    # Determine final number of attempts for reporting
    # If loop finished by verification success (success=True), attempts = attempts made
    # If loop finished by failure (budget, max_attempts, no_change_needed, error),
    # the number of attempts *initiated* is 'attempts + 1' unless max_attempts was exactly hit.
    # The tests seem to expect the number of attempts *initiated*.
    # Let's refine the calculation slightly for clarity.
    # 'attempts' holds the count of *completed* loops (0-indexed).
    # 'current_attempt' holds the user-facing number (1-indexed) of the loop *currently running or just finished*.
    final_attempts_reported = attempts
    if not success:
        # If failure occurred, it happened *during* or *after* the 'current_attempt' was initiated.
        # If loop broke due to budget/no_change/error, current_attempt reflects the attempt number where failure occurred.
        # If loop broke because attempts >= max_attempts, the last valid value for current_attempt was max_attempts.
        # The number of attempts *tried* is current_attempt.
        # However, the tests seem aligned with the previous logic. Let's stick to it unless further tests fail.
        final_attempts_reported = attempts if success else (attempts + 1 if attempts < max_attempts and process.returncode != 0 else attempts)
        # Re-evaluating the test logic:
        # - Budget test: attempts=1 when loop breaks, expects 2. (attempts+1) -> 2. Correct.
        # - Max attempts test: attempts=0 when loop breaks (no change), max_attempts=2, expects <=2. (attempts+1) -> 1. Correct.
        # - If max_attempts=2 was reached *normally* (failed attempt 1, failed attempt 2), attempts would be 2.
        #   The logic `attempts + 1 if attempts < max_attempts else attempts` would return 2. Correct.
        # Let's simplify the return calculation based on 'attempts' which counts completed loops.
        final_attempts_reported = attempts # Number of fully completed fix cycles
        if not success and process and process.returncode != 0: # If we failed after at least one verification run
             # Count the final failed attempt unless success was achieved on the very last possible attempt
             if attempts < max_attempts:
                 final_attempts_reported += 1


    return (
        success,
        final_program_content,
        final_code_content,
        final_attempts_reported, # Use the refined calculation
        total_cost,
        model_name,
    )

# Example usage (requires a dummy fix_code_module_errors and verification script)
# (Keep the example usage block as is for demonstration/manual testing)
if __name__ == "__main__":
    # Create dummy files for demonstration
    DUMMY_CODE_FILE = "dummy_code.py"
    DUMMY_VERIFICATION_FILE = "dummy_verify.py"
    DUMMY_ERROR_LOG = "dummy_error.log"

    # Dummy code with an error
    Path(DUMMY_CODE_FILE).write_text(
        "def my_func(a, b):\n    return a + b # Potential type error if strings used\n",
        encoding='utf-8'
    )

    # Dummy verification script that will fail initially
    Path(DUMMY_VERIFICATION_FILE).write_text(
        f"""
import sys
# Import the function from the code file
try:
    # Assume dummy_code.py is in the same directory
    from dummy_code import my_func
except ImportError as e:
    print(f"Import Error: {{e}}", file=sys.stderr)
    sys.exit(1)

# This will cause a TypeError initially
try:
    result = my_func(5, "a") # Intentionally cause error
    print(f"Result: {{result}}")
    # Check if result is as expected (it won't be initially)
    # Add more checks if needed
    # if result != expected_value:
    #    print(f"Assertion failed: Result {{result}} != expected_value", file=sys.stderr)
    #    sys.exit(1)
except Exception as e:
    print(f"Runtime Error: {{e}}", file=sys.stderr)
    sys.exit(1) # Exit with non-zero code on error

# If we reach here, it means no exceptions occurred
print("Verification passed.")
sys.exit(0) # Exit with zero code for success
""",
        encoding='utf-8'
    )

    # Dummy fix_code_module_errors function (replace with actual import)
    # This dummy version simulates fixing the code on the second attempt
    _fix_attempt_counter = 0
    def dummy_fix_code_module_errors(program, prompt, code, errors, strength, temperature, verbose):
        global _fix_attempt_counter
        _fix_attempt_counter += 1
        cost = 0.05 # Simulate API cost
        model = "dummy-fixer-model-v1"
        analysis = f"Analysis based on errors (attempt {_fix_attempt_counter}):\n{errors[-200:]}" # Show recent history

        if _fix_attempt_counter >= 2:
             # Simulate fixing the code file on the second try
             fixed_code = "def my_func(a, b):\n    # Fixed: Ensure inputs are numbers or handle types\n    try:\n        return float(a) + float(b)\n    except (ValueError, TypeError):\n        return 'Error: Invalid input types'\n"
             # Simulate fixing the verification program to use valid inputs
             fixed_program = program.replace('my_func(5, "a")', 'my_func(5, 10)') # Fix the call
             return True, True, fixed_program, fixed_code, analysis, cost, model # update_program, update_code
        else:
             # Simulate no changes needed on the first try, but still return cost
             return False, False, program, code, analysis + "\nNo changes suggested this time.", cost, model

    # Replace the actual import with the dummy for this example run
    original_fix_func = fix_code_module_errors
    fix_code_module_errors = dummy_fix_code_module_errors

    rprint("[bold yellow]Running example fix_code_loop...[/bold yellow]")

    results = fix_code_loop(
        code_file=DUMMY_CODE_FILE,
        prompt="Create a function that adds two numbers.",
        verification_program=DUMMY_VERIFICATION_FILE,
        strength=0.5,
        temperature=0.1,
        max_attempts=3,
        budget=1.0,
        error_log_file=DUMMY_ERROR_LOG,
        verbose=True,
    )

    rprint("\n[bold blue]----- Fix Loop Results -----[/bold blue]")
    rprint(f"Success: {results[0]}")
    rprint(f"Total Attempts Reported: {results[3]}") # Updated label
    rprint(f"Total Cost: ${results[4]:.4f}")
    rprint(f"Model Name: {results[5]}")
    if results[0]: # Only print final code/program if successful
        rprint("\nFinal Code:")
        rprint(f"[code]{results[2]}[/code]")
        rprint("\nFinal Verification Program:")
        rprint(f"[code]{results[1]}[/code]")
    else:
        rprint("\nFinal Code: [Not successful, code not returned]")
        rprint("Final Verification Program: [Not successful, program not returned]")


    rprint(f"\nCheck the error log file: {DUMMY_ERROR_LOG}")
    if Path(DUMMY_ERROR_LOG).exists():
        rprint("\n[bold blue]----- Error Log Content ----- [/bold blue]")
        log_content = Path(DUMMY_ERROR_LOG).read_text(encoding='utf-8')
        # Use Rich Panel or just print for log content display
        from rich.panel import Panel
        rprint(Panel(log_content, title=DUMMY_ERROR_LOG, border_style="dim blue"))


    # Restore original function if needed elsewhere
    fix_code_module_errors = original_fix_func

    # Clean up dummy files
    # try:
    #     os.remove(DUMMY_CODE_FILE)
    #     os.remove(DUMMY_VERIFICATION_FILE)
    #     # Keep the log file for inspection
    #     # os.remove(DUMMY_ERROR_LOG)
    #     # Remove backups if they exist
    #     for f in Path(".").glob("dummy_*_original_backup.py"): os.remove(f)
    #     for f in Path(".").glob("dummy_code_*.py"): # Remove attempt backups like dummy_code_1.py
    #          if "_original_backup" not in f.name: os.remove(f)
    #     for f in Path(".").glob("dummy_verify_*.py"): # Remove attempt backups like dummy_verify_1.py
    #          if "_original_backup" not in f.name: os.remove(f)
    # except OSError as e:
    #     print(f"Error cleaning up dummy files: {e}")